/**************************************************
 *
 * \author Kristin Reiche
 *
 **************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <math.h>
#include <sys/stat.h>
#include <errno.h>

/* RNAsoup stuff */
#include "RNAsoup.h"
#include "binTree.h"
#include "Utils.h"

#define VERSION "Version 1.0a"

/**************************************************
 *
 * main
 * \param int argc, char** argv
 **************************************************/
int main(int argc, char** argv)
{

  FILE *tree_file=NULL, *tree_eval=NULL, *partition_file=NULL;
  DIR* dirptr;
  char* fasta=NULL;     /* Input file */
  //char* distFile=NULL;  /* Pairwise distance list */
  char* mfeFile=NULL;   /* File holding for each mLocARNA intermediate alignment the RNAalifold mfe */
  char* outdir=NULL;    /* General output directory */
  char foutname[1024];
  char outnodes[1024];  /* Output directory for FASTA and guide tree files */
  char message[1024];   /* For error messages */
  int i, r, N,j;
  int m, t=0, cntParams=0, userDefinedK=0;
  double k;
  float cutoff=-1.;
  struct tree btree;    /* The clustering tree */
  struct node** leaves;
  /************ Command line parameters **************/
  if(argc == 1) exit_with_help();
  for(i=1;i<argc;i++){
    if(argv[i][0] == '-'){
      switch ( argv[i][1] ){
      case 't':
        tree_file = fopen(argv[++i], "r");
        if(tree_file == NULL) {
          fprintf(stderr, "ERROR: Can't open tree '%s'\n", argv[i]);
          exit_with_help();
        }
	cntParams++;
        break;
      case 'f':                                /* FASTA file of sequences. 
                                                  Seq. must be given on one line! */
	fasta = argv[++i];
	cntParams++;
	break;
      case 'm':                                /* File containing the MFEs of consensus structures */
	mfeFile = argv[++i];
	cntParams++;
	break;
      case 'o':
	outdir = argv[++i];
	cntParams++;
	break;
      case 'k':      // single cutoff
	r = sscanf(argv[++i], "%f", &cutoff);
	if(r!=1) exit_with_help();
	cntParams++;
	userDefinedK=1;
        if(cutoff<0){ fprintf(stderr, "ERROR: k must be > 0.\n"); exit_with_help(); }
	break;
      case 'h':
	exit_with_help();
	break;
      case 'v':
	print_version();
	return 1;
	break;
      default: exit_with_help();
      }
    }
  }

  if(cntParams != 5 && cntParams != 4) exit_with_help();
  /* Does output directory already exists? 
     Otherwise create it with according subdirectories. */
  if((dirptr=opendir(outdir))!=NULL){
    //fprintf(stderr,"\n\nERROR: Output directory '%s' already exists!\n\n",outdir);
    closedir(dirptr);
    //return 0;
  }
  else{
    errno=0;
    mkdir(outdir,S_IRWXU | S_IRGRP | S_IXGRP | S_IROTH | S_IXOTH);
    if(errno!=0){
      sprintf(message,"ERROR while 'mkdir %s'",outdir);
      perror("message");
      return 0;
    }
    sprintf(outnodes,"%s/NODES",outdir);
    errno=0;
  }
  /************ Get tree **********************/
  fprintf(stderr, "FILL NAMES AND SEQ ...\n");  /* MUST be filled before calling read_tree! Because, the indices for leaves 
                                                   in NAMES and SEQ are set by searching corresponding name in NAMES. */
  fill_arrays(fasta);
  fprintf(stderr,"READ TREE ...\n");
  if(!read_tree(tree_file,&btree)){ 
	fclose(tree_file); exit(0);
  }
  else fclose(tree_file);
  //fprintf(stderr,"FILL DISTANCE MATRIX ...\n");
  //fill_distance_matrix(distFile);
  fprintf(stderr,"SET NUMBER LEAVES ...\n");
  set_no_leaves(&btree.root);
  fprintf(stderr,"SET HEIGHT ...\n");
  set_height(&btree.root);
  fprintf(stderr,"SET CONSENSUS MFE and SCI ...\n"); /* single MFE is set while reading the tree */
  set_consensus_mfe(&btree.root, mfeFile);

  // print tree, deprecated cause it's nearly the same tree as in RNAclust-folder
  //sprintf(foutname,"%s/tree", outdir);
  //if((tree_eval=fopen(foutname,"w")) == NULL){
  //  sprintf(message,"ERROR while open file '%s' for writing", tree_eval);
  //  perror(message);
  //  return 0;
  //}
  //print_tree(&btree.root, tree_eval);  
  //fclose(tree_eval);

  /* Test distances */
  /*N=0;
  get_leaves(&btree.root,&leaves,&N);
  for(i=0; leaves[i]!=NULL; i++){
    for(j=i+1; leaves[j]!=NULL; j++){
      float dist = get_distance(leaves[i]->element.idx, leaves[j]->element.idx);
      if(leaves[i]->element.idx > leaves[j]->element.idx){
	fprintf(stderr, "%d %d %g\n", leaves[i]->element.idx+1, leaves[j]->element.idx+1, dist);
	//fprintf(stderr, "%s %s %g\n", get_name(leaves[i]->element.idx), get_name(leaves[j]->element.idx));
      }
      else{
	fprintf(stderr, "%d %d %g\n", leaves[j]->element.idx+1, leaves[i]->element.idx+1, dist);
	//fprintf(stderr, "%s %s %g\n", get_name(leaves[j]->element.idx), get_name(leaves[i]->element.idx));
      }
    }
  }
  return;  
  */

  /************ Optimal partition **************/
  fprintf(stderr,"FIND BEST PARTITION ...\n");
  if(userDefinedK) ASIGLEVEL[0] = cutoff;
  else{
	  for(m=0, k=0.; m<N_SIG_LEVELS; m++, k+=0.2){
	    ASIGLEVEL[m]=k;
  	  }
  }
  partition_consmfe(&btree.root, cutoff);
  fprintf(stderr,"PARTITION END ...\n");

  fprintf(stderr,"PRINT PARTITIONS ...\n");
//  if(userDefinedK){
//	sprintf(foutname, "%s/partition_k%.2f.txt", outdir, cutoff);
//	if((partition_file=fopen(foutname,"w")) == NULL){
//                sprintf(message,"ERROR while open file '%s' for writing", foutname);
//                perror(message);
//                return 0;
//        }
//	print_partition(&btree.root, 0, partition_file);
//        fclose(partition_file);
//  }
//  else
//  {
//    for(m=0, k=0.; m<N_SIG_LEVELS; m++, k+=0.2){
//	sprintf(foutname, "%s/partition_k%.2f.txt", outdir, k);
//	if((partition_file=fopen(foutname,"w")) == NULL){
//  		sprintf(message,"ERROR while open file '%s' for writing", foutname);
//		perror(message);
//		return 0;
//	}
//	print_partition(&btree.root, m, partition_file);
//	fclose(partition_file);
//    }
//  }
// print tree
  if(!userDefinedK){
	  sprintf(foutname,"%s/tree_xml", outdir);
	  if((tree_eval=fopen(foutname,"w")) == NULL){
	    sprintf(message,"ERROR while open file '%s' for writing", foutname);
	    perror(message);
	    return 0;
	  }
	  fprintf(tree_eval,"<?xml version=\"1.0\"?>\n");
	  print_tree_xml(&btree.root, tree_eval);  
	  fclose(tree_eval);
   }

  /************ Free everything **************/
  free_all(&btree);
  
  return 1;
  
}

void free_all(struct tree *btree)
{
  fprintf(stderr,"FREE TREE ...\n");
  free_tree(&(btree->root));
  fprintf(stderr,"FREE SEQ ARRAY ...\n");
  utils_free_mcarray(SEQ);
  fprintf(stderr,"FREE NAMES ARRAY ...\n");
  utils_free_mcarray(NAMES);
  fprintf(stderr,"FREE DIST ARRAY ...\n");
  free(DIST);
  return;
}

/*********************************************
 *
 * Partition according to difference to cons_mfe.
 *
 *********************************************/
void partition_consmfe(struct node* anode, double alpha)
{
  if(!anode->isLeaf){   // internal node? 
    
    partition_consmfe(anode->left, alpha);
    partition_consmfe(anode->right, alpha);

    int i;
    double sum_all = get_consmfe_sum(anode->left, anode->cons_mfe) + get_consmfe_sum(anode->right, anode->cons_mfe);
    double sum_l = get_consmfe_sum(anode->left, anode->left->cons_mfe);
    double sum_r = get_consmfe_sum(anode->right, anode->right->cons_mfe);

    if(sum_all != 0) anode->testValue = (double)(sum_l+sum_r)/sum_all;
    else anode->testValue = 1;    // because if sum_all=0, then sum_l and sum_r should also be almost 0
    
    for(i=0; i<N_SIG_LEVELS;i++){ 

      double critValue = utils_get_duda_critical_value(ASIGLEVEL[i],anode->noLeaves);   
      if(anode->testValue < critValue || (fabs(anode->cons_mfe-0.)<1e-6 && fabs(anode->right->cons_mfe-0.)<1e-6 && fabs(anode->left->cons_mfe-0.)<1e-6)){   // reject null hypothesis that seq. belong to one cluster
	anode->isGroup_consmfe[i] = 0;
      }
      else{
	// Parent node can only be seen as a group in case children are also groups or leaves
	if(anode->left->isGroup_consmfe[i] && anode->right->isGroup_consmfe[i])  
	  anode->isGroup_consmfe[i] = 1;
	else
	  anode->isGroup_consmfe[i] = 0;
      }
    }
  }
  else{
    int i;
    // leaves are always a singleton, no matter of significance level
    for(i=0; i<N_SIG_LEVELS;i++) anode->isGroup_consmfe[i] = 1;  
  }
  return;
}

double get_consmfe_sum(struct node* anode, double cons_mfe)
{
  struct node** leaves;
  int N=0, i;
  double sum;

  /* some checks before */
  if(cons_mfe > 0){
    fprintf(stderr, "ERROR (get_consmfe_sum): cons_mfe > 0 (%f)!\n", cons_mfe);
    print_info_node(anode, stderr, 1);
    exit(0);
  }

  get_leaves(anode, &leaves, &N);

  sum = 0.; 
  for(i=0; leaves[i]!=NULL; i++){
    if(leaves[i]->single_mfe > 0){
      fprintf(stderr, "ERROR (get_consmfe_sum): single_mfe > 0 (%f %s)!\n", leaves[i]->single_mfe, get_name(leaves[i]->element.idx));
      exit(0);
    }
    double x = pow((leaves[i]->single_mfe - cons_mfe), 2.);  // this is the random variable of interest, i.e. sqared distance to cons_mfe
    sum += (x/1000.);           // to avoid overflow (x/1000.)
  }

  free(leaves);

  return sum;

}

double avg_mfe(struct node* anode)
{
  struct node** leaves;
  int N=0, i;
  double sum;

  get_leaves(anode, &leaves, &N);

  sum = 0.;
  for(i=0; leaves[i]!=NULL; i++){
    if(leaves[i]->single_mfe > 0){
      fprintf(stderr, "ERROR (get_consmfe_sum): single_mfe > 0 (%f %s)!\n", leaves[i]->single_mfe, get_name(leaves[i]->element.idx));
      exit(0);
    }
    sum += leaves[i]->single_mfe;           // to avoid overflow (x/1000.)
  }

  free(leaves);

  return (double)sum/(double)anode->noLeaves;

}

void print_partition(struct node* anode, int idx, FILE* stream)
{
  if(anode->isGroup_consmfe[idx]){
    print_info_node(anode, stream, 1);
  }
  else{
    if(!anode->isLeaf){
      print_partition(anode->left, idx, stream);
      print_partition(anode->right, idx, stream);
    }
  }
  return;
}

/**************************************************
 *
 * 
 **************************************************/

void print_version()
{
  printf("%s\n", VERSION);
}

void exit_with_help()
{
  printf("\nUsage:\tRNAsoup [-t tree] [-f fasta] [-m mfe_consensus] [-o outdir] [-k num] [-h] [-v]\n\n"
	 "-t file     Tree in NEWICK format\n"
	 "-f file     FASTA file of all sequences in tree\n"
         "-m file     File containing the RNAalifold consensus MFE for each subtree < 500 sequences\n"
	 "-o dir      Output directory which is created to store the output\n"
	 "-k float    Significance level k\n"
	 "-h          Show this help message\n"
	 "-v          Print version information\n"
	 "\n");
  exit(0);
}
