/**************************************************
 *
 * \author Kristin Reiche
 *
 **************************************************/


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <dirent.h>
#include <errno.h>

#include "binTree.h"
#include "Utils.h"

#define MIN(x,y)       (((x)<(y)) ? (x) : (y))
#define MAX(x,y)       (((x)>(y)) ? (x) : (y))

// Global variables
struct node nodes[100];

/**************************************************
 * Read tree from file
 * \return int (1 on success, 0 otherwise)
 **************************************************/
int read_tree(
	      FILE *file, 
	      struct tree* atree)
{
  char *line=NULL, *ptr=NULL;
  int n=0,i=0,pos=0,id=0;
  int length;
  
  init_node(&(atree->root),
	    (struct node*)NULL,(struct node*)NULL,(struct node*)NULL,0,id,0);
  atree->noLeaves=0;
  
  while((line=utils_getline(file))!=NULL){

    /* Test for semicolon at end of tree */
    length = strlen(line);
    if(*(line+length-1) != ';'){
      fprintf(stderr, 
	      "ERROR: Last character of input tree must be ';' and not '%s'!\n",
	      line+length-2);
      free(line);
      return 0;
    }

    /* White space in input string? */
    ptr=line;
    while(*ptr != '\0'){
      if(strncmp(ptr," ",1)==0){
	fprintf(stderr, "ERROR: No white space allowed!\n");
	free(line);
	return 0;
      }
      ptr++;
    }

    /* Built binary tree */
    ptr=line;
    ptr++; /* Ignore first open bracket as it spans the whole tree and not a subtree.
	      But brackets are assumed to introduce a subtree!
	      If ptr would not be increased the real root is found at root->left! */
    /* Level is 1 as root is already at level 0 */
    built_tree(&ptr,&(atree->root),1,0,&(atree->noLeaves),&id);
    /* Check number of leaves. Last entry must be a NULL pointer. */
    if(atree->noLeaves >= MAX_NO_LEAVES){
      fprintf(stderr,
	      "ERROR: No. leaves is (%d), but maximal number is ",
	      atree->noLeaves);
      fprintf(stderr,"restricted to %d!\n",MAX_NO_LEAVES);
      free_tree(&(atree->root));
      free(line);
      return 0;
    }
    NLEAVES = atree->noLeaves;
    n++;
    free(line);
  }

  if(n>1){
    /* Tree must be given on one line in NEWICK format. */
    fprintf(stderr,"ERROR: Tree must be given as one line in NEWICK format!\n");
    free_tree(&(atree->root));
    return 0;
  }

  return 1;

}

/**************************************************
 * Recursive routine which builts tree
 * \return void
 **************************************************/
void built_tree(
		char** sptr,            /* Pointer-pointer to string */
		struct node* anode,     /* Node */
		int level,              /* Current level of tree */
		int lastLeaf,           /* Last node was a leaf? */
		int* cntLeaves,         /* Ptr to int, counts the number of leaves */
		int* id                 /* Ptr to int, counts the number of nodes */
		)
{

  //fprintf(stderr,"%d: n'%s %s %d'\n",h, *sptr, *sptr-1,&(**sptr));
  int stop=0;
  while(!stop){
    if(strncmp(*sptr,"(",1)==0 && strncmp(*sptr-1,",",1)!=0){
      anode->left = (struct node*)calloc(1,sizeof(struct node));
      (*id)++;
      init_node(anode->left, 
		anode,(struct node*)NULL,(struct node*)NULL,level,*id,0);
      (*sptr)++; 
      built_tree(sptr,anode->left,level+1,0,cntLeaves,id);
      anode->left->branch_length = get_branch_length(sptr);
      //fprintf(stderr,"%d: ('%s %s %d'\n",h, *sptr, *sptr-1,&(**sptr));
    }
    if(strncmp(*sptr,"(",1)==0 && strncmp(*sptr-1,",",1)==0){
      anode->right = (struct node*)calloc(1,sizeof(struct node));
      (*id)++;
      init_node(anode->right, 
		anode,(struct node*)NULL,(struct node*)NULL,level,*id,0);
      (*sptr)++; 
      built_tree(sptr,anode->right,level+1,0,cntLeaves,id);
      anode->right->branch_length = get_branch_length(sptr);
      //fprintf(stderr,"%d: (,'%s %s %d'\n",h, *sptr, *sptr-1,&(**sptr));
    }
    // left leaf
    if(is_letter(*sptr)==1 && strncmp(*sptr-1,"(",1)==0 && lastLeaf==0){
      anode->left = (struct node*)calloc(1,sizeof(struct node));
      (*id)++;
      init_node(anode->left, 
		anode, (struct node*)NULL, (struct node*)NULL,level,*id,1);
      built_tree(sptr,anode->left,level+1,1,cntLeaves,id);
      anode->left->branch_length = get_branch_length(sptr);
      //fprintf(stderr,"%d: ll'%s %s %d'\n",h, *sptr, *sptr-1,&(**sptr));
    }
    // right leaf
    if(is_letter(*sptr)==1 && strncmp(*sptr-1,",",1)==0 && lastLeaf==0){
      anode->right = (struct node*)calloc(1,sizeof(struct node));
      (*id)++;
      init_node(anode->right, 
		anode, (struct node*)NULL, (struct node*)NULL,level,*id,1);
      built_tree(sptr,anode->right,level+1,1,cntLeaves,id);
      anode->right->branch_length = get_branch_length(sptr);
      //fprintf(stderr,"%d: lr'%s %s %d'\n",h, *sptr, *sptr-1,&(**sptr));
    }
    // name of leaf
    if(is_letter(*sptr)==1 && lastLeaf==1){
      built_leaf(sptr,anode);		
      (*sptr)++;
      (*cntLeaves)++;
      //fprintf(stderr,"%d: ln'%s %s %d'\n",h,*sptr, *sptr-1,&(**sptr));
      return;	
    }
    if(strncmp(*sptr,",",1)==0){
      (*sptr)++;
      //fprintf(stderr,"%d: ,'%s %s %d'\n",h, *sptr, *sptr-1,&(**sptr));
    }
    if(strncmp(*sptr,")",1)==0){
      (*sptr)++;
      return;
    }
    if(**sptr==';'){
      stop=1;
    }
  }
  return;
}

/**************************************************
 * Set name of a leaf and find corresponding index 
 * in global arrays NAMES. Gives then also index in 
 * global arrays SEQ and DIST.
 * \return void
 **************************************************/
void built_leaf(
		char** str,             /* Pointer-pointer to current char */
		struct node* anode)     /* Leaf */
{
  int l=0,i;
  char* ptr= *str;
  char* name;
  char* seq;
  char* struc;
  char message[1024];

  while(is_letter(ptr)){
    l++; ptr++;
  }
  name = (char*)calloc(l+1,sizeof(char));
  strncpy(name, *str, l);   

  /* Now search for this name in global array NAMES in order 
     to retrieve index for this leaf. */
  for(i=0; NAMES[i]!=NULL; i++){
    if(strcmp(name, NAMES[i])==0){
      anode->element.idx = i;
    }
  }

  if(anode->element.idx == -1){
    sprintf(message,"ERROR while builting leaf: No name found in array NAMES ('%s').",name);
    perror(message);
    return;
  }

  /* Now set mfe of single sequence as we know the index of the leaf */
  seq = get_seq(anode->element.idx);
  anode->single_mfe = utils_fold_seq(seq);
  
  (*str)+=(l-1);
  return;
}

/**************************************************
 * Get branch length for current node
 * \return float
 **************************************************/
float get_branch_length(char** str)   /* Pointer-pointer to current char */
{
  float bl=-1;
  int l=0;
  char* ptr= *str;
  char* str_bl=NULL;

  // Is there a colon which defines start point for branch length?
  if(strncmp(ptr,":",1)==0){
    ptr++;
  }
  else{
    fprintf(stderr,"ERROR: No colon for branch length found!\n");
    exit(0);
  }
     
  // Get branch length
  while(is_letter(ptr)){
    l++; ptr++;
  }
  str_bl = (char*)calloc(l+1,sizeof(char));
  strncpy(str_bl,(*str+1),l);
  sscanf(str_bl,"%g",&bl);
  
  // Set pointer just after branch length
  (*str)+=(l+1);
  
  free(str_bl);
  return bl;
  
}

/**************************************************
 * Initialize a node 
 * \return void
 **************************************************/
void init_node(
	       struct node* anode,   /* Pointer to current node */
	       struct node* parent,  /* Pointer to parent of current node */
	       struct node* left,    /* Pointer to left child of current node */
	       struct node* right,   /* Pointer to right child of current node */ 
	       int level,            /* Level of current node */
	       int id,               /* Unique ID of current node */
	       int isLeaf)           /* Identifier if current node is a leaf */
{
  anode->parent = parent;
  anode->left = left;
  anode->right = right;
  anode->level = level;
  anode->cons_mfe = 1;
  anode->single_mfe = 1;
  anode->height = -1.;
  anode->branch_length = -1.;
  anode->isLeaf = isLeaf;
  anode->element.id = id;
  anode->element.idx = -1.;
  anode->element.clustalw = (char*) NULL;
  anode->element.locarna = (char*) NULL;
  anode->noLeaves=-1;
  anode->testValue=-1.;
  return;
}

/**************************************************
 * Set for each node the height 
 * branch_length must be set!!
 * \return int
 **************************************************/
float set_height(struct node* anode)
{
  if((anode->left == NULL || anode->right == NULL) && !anode->isLeaf){
    fprintf(stderr,"ERROR: Node '%s' is not a leaf but has children ",
	    get_name(anode->element.idx));
    fprintf(stderr,"to null pointer (set_height)!\n");
    return 0;
  }
  
  if(!anode->isLeaf){	/* internal node? */
    float h1,h2;
    h1 = set_height(anode->left);
    h2 = set_height(anode->right);
    if(fabs(h1-h2)>0.00001){
      //fprintf(stderr,"WARNING: Heights of children do not equal!\n");
    }
    anode->height = h1 + anode->left->branch_length;      
    if(h2 + anode->right->branch_length > anode->height)
      anode->height = h2 + anode->right->branch_length;
  }
  else{
    if(anode->branch_length==-1){
      fprintf(stderr,"ERROR: No branch length!\n");
      return 0;
    }
    anode->height = 0; //anode->branch_length;
  }
  return anode->height;
}


/**************************************************
 * Set number of leaves. 
 * \return int
 **************************************************/
int set_no_leaves(struct node* anode)
{
  char message[1024];
  
  if((anode->left == NULL || anode->right == NULL) && !anode->isLeaf){
    fprintf(stderr,"ERROR: Node '%s' is not a leaf but has children ",
	    get_name(anode->element.idx));
    fprintf(stderr,"to null pointer (set_no_leaves)!\n");
    return 0;
  }
  
  if(!anode->isLeaf){	/* internal node? */
    struct node** leaves;
    int N=0;

    set_no_leaves(anode->left);
    set_no_leaves(anode->right);
    get_leaves(anode,&leaves,&N);
    anode->noLeaves=N;

    if(anode->noLeaves != (anode->left->noLeaves + anode->right->noLeaves)){
      fprintf(stderr,
	      "ERROR (set_no_leaves): Node %d has %d leaves, but children sum to %d\n",
	      anode->element.id, anode->noLeaves, anode->left->noLeaves,
	      anode->right->noLeaves);
      return 0;
    }

    free(leaves);
  }
  else{
    anode->noLeaves=1;
  }
  return 1;
}


/*******************************************************************
 * Fill global dynamic arrays NAMES and SEQ
 * Use orderFASTAlikeNAMES.pl to create the FASTA file.
 * \returns Number of names
 *******************************************************************/
int fill_arrays(char *fastaFile)
{
  char *line = NULL;
  char message[1024];
  int i=0;
  FILE *fpin;

  /* Open names file and save names in array names */
  if((fpin=fopen(fastaFile,"r")) == NULL){
    sprintf(message,"ERROR while open FASTA file '%s' for reading",fastaFile);
    perror(message);
    return 0;
  }

  i=0;
  NAMES = calloc(1,sizeof(char*));
  SEQ = calloc(1,sizeof(char*));
  while((line=utils_getline(fpin))!=NULL){
    if(strncmp(line, ">", 1)==0){
      char name[1024];
      sscanf(line+1,"%s", &name);
      NAMES[i] = strdup(name);
      char** tmp = realloc(NAMES, (i+2)*sizeof(char*));
      if(tmp != NULL){
	NAMES = tmp;
	tmp = NULL;
      }
      else{
	sprintf(message,"ERROR while reading FASTA file: realloc on NAMES failed ('%s', '%s')",
		fastaFile,line);
	perror(message);
	utils_free_mcarray(NAMES);
	return 0;
      }
      NAMES[i+1] = NULL;
    }
    else if(strncmp(line," ", 1)==0){ 
      free(line); 
      continue; 
    }
    else{
      char seq[1024];
      sscanf(line,"%s", &seq);
      SEQ[i] = strdup(seq);
      char** tmp = realloc(SEQ, (i+2)*sizeof(char*));
      if(tmp != NULL){
	SEQ = tmp;
	tmp = NULL;
      }
      else{
	sprintf(message,"ERROR while reading FASTA file: realloc on SEQ failed ('%s', '%s')",
		fastaFile,line);
	perror(message);
	utils_free_mcarray(SEQ);
	return 0;
      }
      SEQ[i+1] = NULL;
      i++;
    }
    free(line);
  }

  fclose(fpin);
  return i;
}


/*******************************************************************
 * Set for each internal node the MFE of the consensus structure
 * retrieved from RNAalifold.
 *******************************************************************/
void set_consensus_mfe(struct node* anode, char *mfeFile)
{
  if(!anode->isLeaf){

    set_consensus_mfe(anode->left, mfeFile);
    set_consensus_mfe(anode->right, mfeFile);

    set_consensus_mfe_anode(anode, mfeFile);
  }
  else{
    anode->cons_mfe = -1;
  }
  
  return;
}


void set_consensus_mfe_anode(struct node* anode, char *mfeFile)
{
  char *line = NULL;
  char message[1024];
  int i=0;
  FILE *fpin;
  struct node** leaves;
  int N=0, found, checkIt=0;
  char currentAlign[1024];
  int nSeq;
  float avgMFE=0;

  if(anode->isLeaf){
    anode->cons_mfe = 1;
    return;  /* has no consensus structure */
  }

  /* RNAalifold only proceeds up to 500 sequences. clusters, with more than 500 sequences do not have 
     a complete entry in mfeFile, hence discard them and set mfe to 0, because they 
     certainly won't have a consensus structure and energy of nothing is 0. */
  if(anode->noLeaves > 500){
    anode->cons_mfe = 0;
    //fprintf(stderr,">set_cons_mfe: found noLeaves= %d (no RNAalifold call possible)\n", anode->noLeaves);
    return;
  }

  get_leaves(anode, &leaves, &N);

  /* Open mfe file */
  if((fpin=fopen(mfeFile,"r")) == NULL){
    sprintf(message,"ERROR while open mfeFile '%s' for reading",mfeFile);
    perror(message);
    free(leaves);
    return;
  }
    
  /* Process file and look for alignment file of current cluster */
  while((line=utils_getline(fpin))!=NULL){
    
    if(strncmp(line, ">", 1)==0){  /* new alignment */
      sscanf(line+1,"%s",&currentAlign);
      anode->element.locarna = strdup(currentAlign);
      /* Reset everything */
      checkIt = 0;
      found = 0;      
    }
    else if(strncmp(line, "n:", 2)==0){ /* number of sequences in cluster */
      sscanf(line+2,"%d", &nSeq);
      if(nSeq == anode->noLeaves)
	checkIt = 1;
      else
	checkIt = 0;
    }
    else if(strncmp(line, "mfe:", 4)==0){ /* MFE of consensus structure */
      /* Get MFE when we found the cluster */
      if(found == anode->noLeaves && checkIt==1){
	sscanf(line+4,"%f", &anode->cons_mfe);
	//fprintf(stderr,">set_cons_mfe: found %f %d %s %s\n", anode->cons_mfe, anode->noLeaves,line, currentAlign);
	/* Set also sci */
	avgMFE = 0.;
	for(i=0; leaves[i]!=NULL; i++){
	  avgMFE += (float)leaves[i]->single_mfe;
	}
	avgMFE = avgMFE/anode->noLeaves;
	anode->sci = (float)anode->cons_mfe/(float)avgMFE;
	//fprintf(stderr,"sci: %f\n",anode->sci);
	break;
      }
    }
    else{    /* sequences; check them only if number is same as anode->noLeaves  */
      
      if(checkIt){
	char name[1024];
	char *s1, *s2;
	int l, lstr, i;

	s1 = line;
	lstr = strlen(line);
	found = 0;
	while((s2=index(s1,' '))!=NULL){
	  l = s2-s1;  
	  strncpy(name,s1,(size_t)l);
	  name[l]='\0';
	  //fprintf(stderr,"name '%s'\n",name);
	  s1 = s2+1; // for next iteration
	  
	  // Now test if we have such a leaf
	  for(i=0; leaves[i]!=NULL; i++){
	    char* nametree = strdup(get_name(leaves[i]->element.idx));
	    char *s;
	    // names should be identical when getConsMFE.pl is used 
	    //fprintf(stderr,"'%s' '%s' %d\n", name, nametree, l);
	    if(strncmp(name, nametree, l)==0){
	      found++;
	      //fprintf(stderr,"names %s %s\n", name, nametree);
	      break;
	    }
	    free(nametree);
	  }
	}
      }
    }
    
    free(line);

  }/* end while */
  
  if(found != anode->noLeaves){
    fprintf(stderr,">set_cons_mfe: ERROR: alignment not found %d %d\n", found, anode->noLeaves);
    print_info_node(anode,stderr,1);
    exit(0);
  }
  
  free(leaves);
  fclose(fpin);

  return;
}


/*******************************************************************
 * Fill matrix of distances  
 * 
 * A tab delimited file is given as the pairwise distance list 
 * for the leaves of the tree.
 * File format: 1 0 a       
 *              2 0 b
 *              2 1 c
 *              3 0 d
 *              3 1 e
 *              3 2 f
 * with a,b,etc. being distances and 0,1,2 the corresponding indices for 
 * the sequences. Indices MUST mirror the order in FASTA file!
 * The file MUST be given in this format otherwise 
 * distances are fucked up in the program.
 * 
 * binTree handles pairwise distances in one dynamic array:
 *
 * |a|bc|def|ghij|...
 *
 * No negative distances are allowed, as the last element of the array
 * marks with -1 the end of the array!
 *
 * Use orderFASTAlikeNAMES.pl to create the FASTA file.
 *******************************************************************/
/*int fill_distance_matrix(char *distFile)
{
  char message[1024];
  FILE *fpin;
  char *line=NULL;
  int i=0; 
  
  // Open distance file and assign indices for distance matrix 
  if((fpin=fopen(distFile,"r")) == NULL){
    sprintf(message,"ERROR while open distance file '%s' for reading",distFile);
    perror(message);
    return 0;
  }
  line=NULL;
  DIST = calloc(1, sizeof(float));
  while((line=utils_getline(fpin))!=NULL){
    int i1,i2;
    float dist = -1.;
    sscanf(line,"%d %d %f",&i1, &i2, &dist);
    if(dist == -1 || dist < 0){
      sprintf(message,"ERROR while reading distance file '%s': No distance found in line '%s'",distFile,line);
      perror(message);
      return 0;
    }

    // Is smallest index in distance file 1? 
    if(i==0 && i2!=1){
       sprintf(message,"ERROR while reading distance file '%s': Smallest index is not 1!",distFile);
       perror(message);
       return 0;
    }
    i1--; i2--;
    
    // Fill distance array just by pushing the distances as they occur in the input distance file. 
    DIST[i] = dist;
    float* tmp = realloc(DIST, (i+2)*sizeof(float));  
    if(tmp != NULL){
      DIST=tmp;
      tmp = NULL;
    }
    else{
      sprintf(message,"ERROR while reading distance file '%s': realloc failed: %s",distFile,line);
      perror(message);
      free(DIST);
      return 0;
    }
    DIST[i+1] = -1.;
       
    // TEST - Find entry via indices 
    dist = get_distance(i1,i2);
    if(DIST[i] != dist){
      sprintf(message,
	      "ERROR while reading distance file '%s': Wrong indices assigned to leaves - distance is lost: %s",
	      distFile,line);
      perror(message);
      return 0;
    }
    
    free(line);
    i++;
  }

  // Clean everything 
  fclose(fpin);

  return 1;
}
*/

/**************************************************
 * Get leaves of given node anode 
 * \return int
 **************************************************/
int get_leaves(
		struct node* anode,
		struct node*** leafptr, /* Zeiger auf Zeiger auf struct node* Array */
		int* pos)
{
  if((anode->left == NULL || anode->right == NULL) && !anode->isLeaf){
    fprintf(stderr,
	    "ERROR: Node '%s' is not a leaf but has children ",get_name(anode->element.idx));
    fprintf(stderr,"to null pointer (get_leaves)!\n");
    return 0;
  }

  if(!anode->isLeaf){	// internal node?
    get_leaves(anode->left,leafptr,pos);
  }
  if(anode->isLeaf){
    if((*pos)==0){
      (*leafptr) = calloc(1,sizeof(struct node*));
    }
    (*leafptr)[*pos] = anode;
    (*pos)++;
    struct node** tmp = realloc(*leafptr, ((*pos)+1)*sizeof(struct node*));
    if(tmp != NULL){
      *leafptr = tmp;
      tmp = NULL;
    }
    else{
      fprintf(stderr,"ERROR (get_leaves): realloc failed!\n");
      free(leafptr);
    }
    (*leafptr)[*pos] = NULL;
    return 1; 
  }
  if(!anode->isLeaf){   // internal node?
    get_leaves(anode->right,leafptr,pos);
  }
  return 1;
}


/**************************************************
 * Get leaves of given node anode with ID 'id'
 * First call must reference to root node.
 * \return int
 **************************************************/
int get_leaves_id(
		struct node* anode,
		struct node** leafptr,
		int* pos,
		int id,
		int found             /* Node with ID 'id' found? */
		)
{
  if((anode->left == NULL || anode->right == NULL) && !anode->isLeaf){
    fprintf(stderr,
	    "ERROR: Node '%s' is not a leaf but has no children ",get_name(anode->element.idx));
    fprintf(stderr,"to null pointer (get_leaves_id)!\n");
    return 0;
  }

  if(anode->element.id == id){
    found = 1;
  }

  if(!anode->isLeaf){	// internal node?
    get_leaves_id(anode->left,leafptr,pos,id,found);
  }
  if(anode->isLeaf && found==1){   /* Are leaves children of wanted node? */
    leafptr[*pos] = anode;
    (*pos)++;
    leafptr[*pos] = NULL;  /* current end of array */
    return 1;
  }
  if(!anode->isLeaf){   // internal node?
    get_leaves_id(anode->right,leafptr,pos,id,found);
  }
  return 1;
}

/****************************************************************
 * Get name.
 * \return float
 ****************************************************************/
char* get_name(int idx)
{
  if(idx < 0){
    fprintf(stderr, "ERROR (get_name): No valid index '%d'!\n", idx);
    return NULL;
  }
  if(NAMES[idx] == NULL){
    fprintf(stderr,"ERROR (get_name): No name for index '%d'\n",idx);
    return NULL;
  }
  return (NAMES[idx]);
}

/****************************************************************
 * Get sequence.
 * \return float
 ****************************************************************/
char* get_seq(int idx)
{
   if(idx < 0){
    fprintf(stderr, "ERROR (get_seq): No valid index '%d'!\n", idx);
    return NULL;
   }
  if(SEQ[idx] == NULL){
    fprintf(stderr,"ERROR (get_seq): No sequence for index '%d'\n",idx);
    return NULL;
  }
  return (SEQ[idx]);
}

/****************************************************************
 * Get pairwise distance
 * \return float
 ****************************************************************/
/*float get_distance(int idx1, int idx2)
{
  int sum=0, i;

  if(idx1 < idx2){
    int t = idx2;
    idx2 = idx1;
    idx1 = t;
  }
  for(i=0; i<idx1;i++){
    sum += i;
  }
  if( sum+idx2 > ((((NLEAVES-1)*NLEAVES)/2)-1) ){
    fprintf(stderr, "ERROR: No distance found for indices '%d' '%d'; NLEAVES=%d\n", idx1, idx2,NLEAVES);
    exit(0);
  }

  return DIST[sum + idx2];
}*/


/**************************************************
 * Write tree to STDOUT
 * \return int
 **************************************************/
int print_tree(
	       struct node* anode, /* Current node */
	       FILE* stream        /* File stream the tree is printed to */
	       )
		
{
  char filename[1024];

  if((anode->left == NULL || anode->right == NULL) && !anode->isLeaf){
    fprintf(stderr,
	    "ERROR: Node '%s' is not a leaf but has no children ",
	    get_name(anode->element.idx));
    fprintf(stderr,"to null pointer (print_tree)!\n");
    return 0;
  }

  if(!anode->isLeaf){	// internal node?
    fprintf(stream,"(");
    print_tree(anode->left, stream);
  }
  if(anode->isLeaf)
    fprintf(stream,"%s",get_name(anode->element.idx));
  if(!anode->isLeaf){   // internal node?
    fprintf(stream,",");
    print_tree(anode->right, stream);
    fprintf(stream,")");
  
    if(anode->element.locarna==NULL){
      fprintf(stderr,"WARNING (print_tree): No alignment file found for node %d (number leaves: %d)!\n",anode->element.id, anode->noLeaves);
      filename[0]='-';
      filename[1]='1';
      filename[2]='\0';
    }
    else{
      char *suffix = strstr(anode->element.locarna, ".aln");
      char *lastSlash = strstr(anode->element.locarna, "/intermediate") + 13;
      int l;
      if(lastSlash == NULL){
	l = &(*suffix) - &(*anode->element.locarna);
	strncpy(filename, anode->element.locarna, (size_t)l);
      }
      else{
	l = &(*suffix) - &(*lastSlash);
	strncpy(filename, lastSlash, (size_t)l);
      }
      filename[l]='\0';
    }
  }

  if(anode->level != 0){              // no branch length for root
    if(anode->isLeaf){
      fprintf(stream,":%g",anode->branch_length);
    }else{
      //fprintf(stream,"p%g:%g",anode->pvalue, anode->branch_length);
      fprintf(stream,"%s:%g",filename, anode->branch_length);
    }
  }
  else{                        
    //fprintf(stream,"p%g:0;\n",anode->pvalue);
    fprintf(stream,"%s:0;\n",filename);
  }
  return 1;

}



/**************************************************
 * Write tree to STDOUT
 * \return int
 **************************************************/
int print_tree_xml(
	       struct node* anode, /* Current node */
	       FILE* stream        /* File stream the tree is printed to */
	       )
		
{
  int i;
  
  if((anode->left == NULL || anode->right == NULL) && !anode->isLeaf){
    fprintf(stderr,
	    "ERROR: Node '%s' is not a leaf but has children ",
	    get_name(anode->element.idx));
    fprintf(stderr,"to null pointer (print_tree_xml)!\n");
    return 0;
  }
  
  if(!anode->isLeaf){	// Internal node
    /* General info */
    fprintf(stream,"<node id=\"%d\" idx=\"%d\">\n", anode->element.id, anode->element.idx);
    fprintf(stream,"\t<nLeaves>%d</nLeaves>\n", anode->noLeaves); 
    fprintf(stream,"\t<mfe>%g</mfe>\n", anode->cons_mfe); 
    fprintf(stream,"\t<sci>%g</sci>\n", anode->sci); 
    fprintf(stream,"\t<child>%d</child>\n", anode->left->element.id); // just for test issues
    fprintf(stream,"\t<child>%d</child>\n", anode->right->element.id); // just for test issues
    if(anode->level != 0){ // no branch length for root 
      fprintf(stream,"\t<branchLength>%g</branchLength>\n", anode->branch_length);
    }
    /* Files */
    if(anode->element.locarna==NULL)
      fprintf(stderr,"WARNING (print_tree_xml): No alignment file found for node %d (number leaves: %d)!\n",anode->element.id, anode->noLeaves);
    else{
      char filename[1024];
      char *suffix = strstr(anode->element.locarna, ".aln");
      char *lastSlash = rindex(anode->element.locarna, '/') + 1;
      int l;
      if(lastSlash == NULL){
	l = &(*suffix) - &(*anode->element.locarna);
     	strncpy(filename, anode->element.locarna, (size_t)l);
      }
      else{
	l = &(*suffix) - &(*lastSlash);
	strncpy(filename, lastSlash, (size_t)l);
      }
      /*fprintf(stderr, "SUFFIX: %s %d LOCARNA: %s %d LENGTH: %d\n", 
	suffix, &(*suffix), anode->element.locarna, &(*anode->element.locarna), l);
      */
      filename[l]='\0';
      fprintf(stream, "\t<files>%s</files>\n", filename);
    }
    /* Group info */
    fprintf(stream, "\t<groupInfo>\n");
    fprintf(stream, "\t\t<testValue>%g</testValue>\n", anode->testValue);
    for(i=0; i<N_SIG_LEVELS; i++){
      double critValue = utils_get_duda_critical_value(ASIGLEVEL[i], anode->noLeaves);
      fprintf(stream, "\t\t<group k=\"%g\" critv=\"%g\">%d</group>\n", ASIGLEVEL[i], critValue, anode->isGroup_consmfe[i]);
    }
    fprintf(stream, "\t</groupInfo>\n");
    print_tree_xml(anode->left, stream);
    print_tree_xml(anode->right, stream);
    fprintf(stream,"</node>\n");
  }
  else{     // Leaf
    fprintf(stream,"<leaf id=\"%d\" name=\"%s\">\n", anode->element.id, get_name(anode->element.idx));
    /* General info */
    fprintf(stream,"\t<mfe>%g</mfe>\n", anode->single_mfe);
    fprintf(stream,"\t<branchLength>%g</branchLength>\n", anode->branch_length);
    /* Files */
    fprintf(stream,"</leaf>\n");
  }
  
  return 1;

}

/**************************************************
 * Prints information about a node.
 * \return int (1 on success, 0 otherwise)
 **************************************************/
void print_info_node(struct node* anode, FILE* stream, int printLeaves)
{
  int i,N=0;
  struct node** leaves;
  
  get_leaves(anode,&leaves,&N);
  fprintf(stream,"============ Node %d ==============\n",anode->element.id);
//  fprintf(stream,"Level:               %d\n",anode->level);
//  fprintf(stream,"Height:              %g\n",anode->height);
//  fprintf(stream,"Index:               %d\n",anode->element.idx);
//  fprintf(stream,"Branch length:       %g\n",anode->branch_length);
  fprintf(stream,"No. leaves:          %d\n", anode->noLeaves);
  fprintf(stream,"consmfe:             %g\n", anode->cons_mfe);
  fprintf(stream,"sci:                 %g\n", anode->sci);
//  if(anode->element.clustalw==NULL)
//    fprintf(stream,"Clustalw:            -\n");
//  else
//    fprintf(stream,"Clustalw:            %s\n",anode->element.clustalw);
  if(anode->element.locarna==NULL)
    fprintf(stream,"Locarna:             -\n");
  else
    fprintf(stream,"Locarna:             %s\n",anode->element.locarna);
  if(printLeaves){
    fprintf(stream,"Leaves:\n",anode->element.id);
    print_leaves(leaves,stream);
  }
  if(anode->left!=NULL && anode->right!=NULL){
    fprintf(stream,"Left child:          %d\n",anode->left->element.id);
    fprintf(stream,"Right child:         %d\n",anode->right->element.id);
  }
  else{
    fprintf(stream,"Left child:          -\n");
    fprintf(stream,"Right child:         -\n");
  }
  fprintf(stream,"===================================\n");

  /* Free everything */
  free(leaves);
}



/**************************************************
 * Prints information about a node with ID 'id'.
 * First call must point to root node.
 * \return int (1 on success, 0 otherwise)
 **************************************************/
void print_info_node_id(struct node* anode, FILE* stream, int printLeaves, int id, int* found)
{
  int i,N=0;

  /* Ready? */
  if((*found)==1){
    return;
  }
  
  if((anode->left == NULL || anode->right == NULL) && !anode->isLeaf){
    fprintf(stderr,
	    "ERROR: Node '%s' is not a leaf but has children ",
	    get_name(anode->element.idx));
    fprintf(stderr,"to null pointer (print_info_node_id)!\n");
    return;
  }

  /* Node found? */
  if(anode->element.id == id){
    print_info_node(anode, stream, printLeaves);
    (*found)=1;
    return;
  }

  if(!anode->isLeaf){	// internal node?
    print_info_node_id(anode->left,stream,printLeaves,id,found);
    print_info_node_id(anode->right,stream,printLeaves,id,found);
  }
  return;
}

/**************************************************
 * Prints leaves.
 * \return int (1 on success, 0 otherwise)
 **************************************************/
void print_leaves(struct node** leafptr, FILE* stream)
{
  int i;
  for(i=0;leafptr[i]!=NULL;i++){
    //fprintf(stream,"leaf %s %g\n",get_name(leafptr[i]->element.idx), leafptr[i]->silwidth_consmfe);
    fprintf(stream,"leaf %s\n",get_name(leafptr[i]->element.idx));
  }
  return;
}

/**************************************************
 * Frees allocated memory 
 * \return void
 **************************************************/
void free_tree(struct node* anode) /* Pointer to current node */
{
  if(anode != NULL){
    free_tree(anode->left);
    free_tree(anode->right);
    free(anode->left);
    free(anode->right);    
    if(anode->isLeaf){
      if(anode->element.clustalw != (char*)NULL)
	free(anode->element.clustalw);
      if(anode->element.locarna != (char*)NULL)
	free(anode->element.locarna);
    }
    anode->left=NULL;
    anode->right=NULL;
    anode->parent=NULL;
    anode=NULL;
  }

  return;
}


/**************************************************
 * Checks if current char belongs to leaf names.
 * \return int (1 on success, 0 otherwise)
 **************************************************/
int is_letter(char* str)
{
  if(strncmp(str,"(",1) != 0 &&
     strncmp(str,")",1) != 0 &&
     strncmp(str,",",1) != 0 &&
     strncmp(str,";",1) != 0 &&
     strncmp(str,":",1) != 0){    
    return 1;
  }
  return 0;
}

