#ifndef TREE_H
#define TREE_H

#include <stdio.h>
#include <stdlib.h>
#include <dirent.h>

/**************************************************
 *
 * Binary tree structures
 *
 **************************************************/
#define MAX_NO_LEAVES 7000 /* Maximal number of leaves the tree can have */
#define N_SIG_LEVELS  10   /* Number of different significance levels of finding groups by partition_consmfe */

double ASIGLEVEL[N_SIG_LEVELS]; /* Array of significance levels for testing (partition_consmfe). Defined in RNAsoup.c. */
char** NAMES;        /* List of names */
char** SEQ;          /* List of sequences */
float* DIST;         /* List of pairwise distances */ 
int NLEAVES;         /* Number of leaves in tree */

struct elem{
  int id;                 /* ID of corresponding node */
  int idx;                /* If node is leaf, then this should be set to a value > 0.
                             Represents index in NAMES, SEQ and DIST. */
  char* clustalw;         /* Path to clustalw alignment */
  char* locarna;          /* Path to locarna alignment */
};

struct node{
  struct node*      parent;
  struct node*      left;
  struct node*      right;
  struct elem       element;

  int level;
  float cons_mfe;          /* only for internal nodes */
  float single_mfe;        /* only for leaves */
  float sci;               /* sci for all nodes */
  float silwidth_consmfe;  /* Silhoutte width in terms of consensus mfe */
  float height;            /* Sum of branch lengths of all subnodes */
  float branch_length;
  int isLeaf;
  int noLeaves;  
  int isGroup_consmfe[N_SIG_LEVELS];   
  double testValue;         /* Test statistics for a certain node to decide
                               if both subcluster belong to same group. */
};

struct tree{
  struct node root;    /* Root node */
  int noLeaves;        /* Number of elements = number of leaves */
};

/**************************************************
 *
 * Input routines for binary trees
 *
 **************************************************/
int read_tree(FILE *file, struct tree* atree);
void built_tree(
		char** str, 
		struct node* anode,
		int h,
		int lastLeaf,
		int* cntLeaves,
		int* id);
void built_leaf(
		char** str,
		struct node* anode);
float get_branch_length(char** str);
void init_node(struct node* anode,
	       struct node* parent,
	       struct node* left,
	       struct node* right,
	       int level,
	       int id,
	       int isLeaf);
float set_height(struct node* anode);
int set_no_leaves(struct node* anode);
int fill_arrays(char* fastaFile);
void set_consensus_mfe(struct node* anode, char *mfeFile);
void set_consensus_mfe_anode(struct node* anode, char *mfeFile);
//int fill_distance_matrix(char* distFile);
/**************************************************
 *
 * Some useful routines.
 *
 **************************************************/
int get_leaves(
		struct node* anode, 
		struct node*** leafptr,
		int* pos);
int get_leaves_id(
		struct node* anode,
		struct node** leafptr,
		int* pos,
		int id,
		int found             /* Node with ID 'id' found? */
		);
float get_mfe_node(struct node* anode);
float get_sci_node(struct node* anode);
char* get_name(int idx);
char* get_seq(int idx);
//float get_distance(int idx1, int idx2);

/**************************************************
 *
 * Output routines for binary trees
 *
 **************************************************/

int print_tree(struct node* anode, 
	       FILE *stream);
int print_tree_xml(struct node* anode, 
	       FILE *stream);
void print_info_node(struct node* anode, 
		     FILE* stream, 
		     int printLeaves);
void print_info_node_id(struct node* anode, 
			FILE* stream, 
			int printLeaves, 
			int id, 
			int* found);
void print_leaves(struct node**, FILE* stream);
void free_tree(struct node* anode);

/**************************************************
 *
 * Helpers
 *
 **************************************************/

int is_letter(char* str);

#endif
